module Parsers # Bukdu.Plug

using ...Deps.HTTP
using ...Deps.URIParser
using ...Deps: Request
using .HTTP.Messages: hasheader, header
using ...Bukdu: Route
using ..Plug
using JSON2

const default_content_parsers = [:json, :urlencoded, :multipart]
content_parsers = Dict{Symbol, Vector{Symbol}}(:default => default_content_parsers)

function Plug.plug(::Module; parsers::Vector{Symbol})
    content_parsers[:default] = parsers
end


# application/x-www-form-urlencoded

struct UrlEncodedScanner
    data::Vector{UInt8}
end

# https://github.com/JuliaWeb/HttpCommon.jl/blob/v0.2.6/src/HttpCommon.jl#L141
function scan(s::UrlEncodedScanner)::Vector{Pair{String,String}}
    query = String(s.data)
    ps = Vector{Pair{String,String}}()
    isempty(query) && return assoc
    for field in split(query, '&'; keepempty=false)
        (k, v) = split(field, '=')
        key = URIParser.unescape_form(k)
        value = URIParser.unescape_form(v)
        push!(ps, Pair(key, value))
    end
    ps
end


# multipart/form-data

mutable struct FormScanner
    data::Vector{UInt8}
    boundary::String
    FormScanner(data::Vector{UInt8}, boundary::String) = new(data, boundary)
end

const CR = 0x0d
const LF = 0x0a
const ContentDisposition = "Content-Disposition: form-data;"

function rstripcr(vec::Vector{UInt8})::Vector{UInt8}
    if !isempty(vec) && vec[end] == CR
        vec[1:end-1]
    else
        vec
    end
end

function rstripcrlf(vec::Vector{UInt8})::Vector{UInt8}
    if !isempty(vec) && vec[end] == LF
        rstripcr(vec[1:end-1])
    else
        vec
    end
end

function content_disposition_fields(io::IOBuffer)::Dict{String,String}
    fields = Dict{String,String}()
    str = String(readuntil(io, LF))
    h = rstripcr(readuntil(io, LF))
    if isempty(h)
    else
        (k, v) = split(String(h), ": ")
        fields[k] = v
        readuntil(io, LF)
    end
    pat = r" (?P<key>[^\"]*)=\"(?P<val>[^\"]*)\""
    for m in eachmatch(pat, str)
        fields[m[:key]] = m[:val]
    end
    fields
end

function scan(s::FormScanner)::Vector{Pair{String,String}}
    ps = Vector{Pair{String, String}}()
    io = IOBuffer(s.data, read=true, truncate=false)
    while !eof(io)
        chunk = readuntil(io, s.boundary)
        if !isempty(chunk)
            chunkio = IOBuffer(chunk)
            skip(chunkio, CR)
            skip(chunkio, LF)
            contentdisposition = readuntil(chunkio, ContentDisposition)
            if !isempty(contentdisposition)
            contentio = IOBuffer(contentdisposition)
            fields = content_disposition_fields(contentio)
            value = String(rstripcrlf(read(contentio)))
            push!(ps, Pair(fields["name"], value))
            end
        end
    end
    ps
end

function fetch_body_params(route::Route, req::Request)::Vector{Pair{String,Any}}
    if hasheader(req.headers, "Content-Type")
        content_type = header(req.headers, "Content-Type")
        request_parsers = content_parsers[:default]
        if :json in request_parsers && "application/json" == content_type
            nt = JSON2.read(String(req.body))
            return [Pair(String(k),v) for (k,v) in pairs(nt)]
        elseif :urlencoded in request_parsers && "application/x-www-form-urlencoded" == content_type
            scanner = UrlEncodedScanner(req.body)
            return scan(scanner)
        elseif :multipart in request_parsers && startswith(content_type, "multipart/form-data")
            pat = r"boundary=\"?([\W\w]*)\"?"
            m = match(pat, content_type)
            if m isa RegexMatch
                boundary = m[1]
                scanner = FormScanner(req.body, string("--", boundary))
                return scan(scanner)
            end
        end
    end
    Vector{Pair{String,Any}}()
end

end # module Bukdu.Plug.Parsers
